// Cliff Cummings Fifo -  CummingsSNUG2002SJ_FIFO1_rev1_2.doc
// http://www.sunburst-design.com/papers/CummingsSNUG2002SJ_FIFO1.pdf

// 05/09/2024 - changed wrst/rrst --> wrst/rrst
// 05/09/2024 - changed name fifo1 --> fifo16
// 05/09/2024 - DATASIZE = 16, ADDRSIZE = 4

// 05/03/2024 - removed the ifdef/else/endif VENDORRAM section


`timescale 1ns / 1ps

`define DATASIZE 16
`define ADDRSIZE 4

module sync_r2w
  (output reg [`ADDRSIZE:0] wq2_rptr,
   input      [`ADDRSIZE:0] rptr,
   input                   wclk, wrst);
  reg [`ADDRSIZE:0] wq1_rptr;
  always @(posedge wclk or posedge wrst)
    if (wrst) {wq2_rptr,wq1_rptr} <= 0;
    else         {wq2_rptr,wq1_rptr} <= {wq1_rptr,rptr};
endmodule

module sync_w2r
  (output reg [`ADDRSIZE:0] rq2_wptr,
   input      [`ADDRSIZE:0] wptr,
   input                   rclk, rrst);
  reg [`ADDRSIZE:0] rq1_wptr;
  always @(posedge rclk or posedge rrst)
    if (rrst) {rq2_wptr,rq1_wptr} <= 0;
    else         {rq2_wptr,rq1_wptr} <= {rq1_wptr,wptr};
endmodule

module fifomem
  (output [`DATASIZE-1:0] rdata,
   input  [`DATASIZE-1:0] wdata,
   input  [`ADDRSIZE-1:0] waddr, raddr,
   input                 wclken, wfull, wclk);


  // RTL Verilog memory model
  localparam DEPTH = 1<<`ADDRSIZE;
  reg [`DATASIZE-1:0] mem [0:DEPTH-1];
  assign rdata = mem[raddr];
  always @(posedge wclk)
    if (wclken && !wfull) mem[waddr] <= wdata;

endmodule

module rptr_empty
  (output reg                rempty,
   output     [`ADDRSIZE-1:0] raddr,
   output reg [`ADDRSIZE  :0] rptr,
   input      [`ADDRSIZE  :0] rq2_wptr,
   input                     rinc, rclk, rrst);
  reg  [`ADDRSIZE:0] rbin;
  wire [`ADDRSIZE:0] rgraynext, rbinnext;
  //-------------------
  // GRAYSTYLE2 pointer
  //-------------------
  always @(posedge rclk or posedge rrst)
    if (rrst) {rbin, rptr} <= 0;
    else         {rbin, rptr} <= {rbinnext, rgraynext};
  // Memory read-address pointer (okay to use binary to address memory)
  assign raddr     = rbin[`ADDRSIZE-1:0];
  assign rbinnext  = rbin + (rinc & ~rempty);
  assign rgraynext = (rbinnext>>1) ^ rbinnext;
  //---------------------------------------------------------------
  // FIFO empty when the next rptr == synchronized wptr or on reset
  //---------------------------------------------------------------
  assign rempty_val = (rgraynext == rq2_wptr);
  always @(posedge rclk or posedge rrst)
    if (rrst) rempty <= 1'b1;
    else         rempty <= rempty_val;
endmodule

module wptr_full
  (output reg                wfull,
   output     [`ADDRSIZE-1:0] waddr,
   output reg [`ADDRSIZE  :0] wptr,
   input      [`ADDRSIZE  :0] wq2_rptr,
   input                     winc, wclk, wrst);
  reg  [`ADDRSIZE:0] wbin;
  wire [`ADDRSIZE:0] wgraynext, wbinnext;
  // GRAYSTYLE2 pointer
  always @(posedge wclk or posedge wrst)
    if (wrst) {wbin, wptr} <= 0;
    else         {wbin, wptr} <= {wbinnext, wgraynext};
  // Memory write-address pointer (okay to use binary to address memory)
  assign waddr = wbin[`ADDRSIZE-1:0];
  assign wbinnext  = wbin + (winc & ~wfull);
  assign wgraynext = (wbinnext>>1) ^ wbinnext;
  //------------------------------------------------------------------
  // Simplified version of the three necessary full-tests:
  // assign wfull_val=((wgnext[`ADDRSIZE]    !=wq2_rptr[`ADDRSIZE]  ) &&
  //                   (wgnext[`ADDRSIZE-1]  !=wq2_rptr[`ADDRSIZE-1]) &&
  //                   (wgnext[`ADDRSIZE-2:0]==wq2_rptr[`ADDRSIZE-2:0]));
  //------------------------------------------------------------------
  assign wfull_val = (wgraynext=={~wq2_rptr[`ADDRSIZE:`ADDRSIZE-1],
                                   wq2_rptr[`ADDRSIZE-2:0]});
  always @(posedge wclk or posedge wrst)
    if (wrst) wfull  <= 1'b0;
    else         wfull  <= wfull_val;
endmodule

module fifo16
  (output [`DATASIZE-1:0] rdata,
   output             wfull,
   output             rempty,
   input  [`DATASIZE-1:0] wdata,
   input              winc, wclk, wrst,
   input              rinc, rclk, rrst);
  wire   [`ADDRSIZE-1:0] waddr, raddr;
  wire   [`ADDRSIZE:0]   wptr, rptr, wq2_rptr, rq2_wptr;
  sync_r2w      sync_r2w  (.wq2_rptr(wq2_rptr), .rptr(rptr),
                           .wclk(wclk), .wrst(wrst));
  sync_w2r      sync_w2r  (.rq2_wptr(rq2_wptr), .wptr(wptr),
                           .rclk(rclk), .rrst(rrst));
  fifomem       fifomem
                          (.rdata(rdata), .wdata(wdata),
                           .waddr(waddr), .raddr(raddr),
                           .wclken(winc), .wfull(wfull),
                           .wclk(wclk));
  rptr_empty    rptr_empty
                          (.rempty(rempty),
                           .raddr(raddr),
                           .rptr(rptr), .rq2_wptr(rq2_wptr),
                           .rinc(rinc), .rclk(rclk),
                           .rrst(rrst));
  wptr_full     wptr_full
                          (.wfull(wfull), .waddr(waddr),
                           .wptr(wptr), .wq2_rptr(wq2_rptr),
                           .winc(winc), .wclk(wclk),
                           .wrst(wrst));
endmodule


